# -*- coding: utf-8 -*-
from gtts.utils import _translate_url
from bs4 import BeautifulSoup
import requests
import logging
import re

__all__ = ['tts_langs']

# Logger
log = logging.getLogger(__name__)
log.addHandler(logging.NullHandler())


def tts_langs(tld="com"):
    """Languages Google Text-to-Speech supports.

    Args:
        tld (string): Top-level domain for the Google Translate host
            to fetch languages from. i.e `https://translate.google.<tld>`.
            Default is ``com``.

    Returns:
        dict: A dictionnary of the type `{ '<lang>': '<name>'}`

            Where `<lang>` is an IETF language tag such as `en` or `pt-br`,
            and `<name>` is the full English name of the language, such as
            `English` or `Portuguese (Brazil)`.

    The dictionnary returned combines languages from two origins:

    - Languages fetched automatically from Google Translate
    - Languages that are undocumented variations that were observed to work and
      present different dialects or accents.

    """
    try:
        langs = dict()
        log.debug("Fetching with '{}' tld".format(tld))
        langs.update(_fetch_langs(tld))
        langs.update(_extra_langs())
        log.debug("langs: {}".format(langs))
        return langs
    except Exception as e:
        raise RuntimeError("Unable to get language list: {}".format(str(e)))


def _fetch_langs(tld="com"):
    """Fetch (scrape) languages from Google Translate.

    Google Translate loads a JavaScript Array of 'languages codes' that can
    be spoken. We intersect this list with all the languages Google Translate
    provides to get the ones that support text-to-speech.

    Args:
        tld (string): Top-level domain for the Google Translate host
            to fetch languages from. i.e `https://translate.google.<tld>`.
            The language names obtained will be in a language locale of the TLD
            (e.g. ``tld=fr`` will retrieve the French names of the languages).
            Default is ``com``.

    Returns:
        dict: A dictionnary of languages from Google Translate

    """
    URL_BASE = _translate_url(tld)

    # The JavaScript file to look for is either:
    # * translate_m.js or
    # * translate_m_<lang-code>.js
    #   e.g. translate_m_fr.js or translate_m_zh-CN.js
    JS_FILE = r'translate_m(|_\S*)\.js'

    # Load HTML
    page = requests.get(URL_BASE)
    soup = BeautifulSoup(page.content, 'html.parser')

    # JavaScript URL
    # The <script src=''> path can change, but not the file.
    # Ex: /zyx/abc/20180211/translate_m.js
    js_path = soup.find(src=re.compile(JS_FILE))['src']
    js_url = "{}/{}".format(URL_BASE, js_path)

    # Load JavaScript
    js_contents = requests.get(js_url).text

    # Approximately extract TTS-enabled language codes
    # RegEx pattern search because minified variables can change.
    # Extra garbage will be dealt with later as we keep languages only.
    # In: "[...]Fv={af:1,ar:1,[...],zh:1,"zh-cn":1,"zh-tw":1}[...]"
    # Out: ['is', '12', [...], 'af', 'ar', [...], 'zh', 'zh-cn', 'zh-tw']
    pattern = r'[{,\"](\w{2}|\w{2}-\w{2,3})(?=:1|\":1)'
    tts_langs = re.findall(pattern, js_contents)

    # Build lang. dict. from main page (JavaScript object populating lang. menu)
    # Filtering with the TTS-enabled languages
    # In: "{code:'auto',name:'Detect language'},{code:'af',name:'Afrikaans'},[...]"
    # re.findall: [('auto', 'Detect language'), ('af', 'Afrikaans'), [...]]
    # Out: {'af': 'Afrikaans', [...]}
    trans_pattern = r"{code:'(?P<lang>.+?[^'])',name:'(?P<name>.+?[^'])'}"
    trans_langs = re.findall(trans_pattern, page.text)
    return {lang: name for lang, name in trans_langs if lang in tts_langs}


def _extra_langs():
    """Define extra languages.

    Returns:
        dict: A dictionnary of extra languages manually defined.

            Variations of the ones fetched by `_fetch_langs`,
            observed to provide different dialects or accents or
            just simply accepted by the Google Translate Text-to-Speech API.

    """
    return {
        # Chinese
        'zh-cn': 'Chinese (Mandarin/China)',
        'zh-tw': 'Chinese (Mandarin/Taiwan)',
        # English
        'en-us': 'English (US)',
        'en-ca': 'English (Canada)',
        'en-uk': 'English (UK)',
        'en-gb': 'English (UK)',
        'en-au': 'English (Australia)',
        'en-gh': 'English (Ghana)',
        'en-in': 'English (India)',
        'en-ie': 'English (Ireland)',
        'en-nz': 'English (New Zealand)',
        'en-ng': 'English (Nigeria)',
        'en-ph': 'English (Philippines)',
        'en-za': 'English (South Africa)',
        'en-tz': 'English (Tanzania)',
        # French
        'fr-ca': 'French (Canada)',
        'fr-fr': 'French (France)',
        # Portuguese
        'pt-br': 'Portuguese (Brazil)',
        'pt-pt': 'Portuguese (Portugal)',
        # Spanish
        'es-es': 'Spanish (Spain)',
        'es-us': 'Spanish (United States)'
    }



# -*- coding: utf-8 -*-
import os
import pytest
from mock import Mock
from six.moves import urllib

from gtts.tts import gTTS, gTTSError
from gtts.lang import _fetch_langs, _extra_langs

# Testing all languages takes some time.
# Set TEST_LANGS envvar to choose languages to test.
#  * 'fetch': Languages fetched from the Web
#  * 'extra': Languagee set in Languages.EXTRA_LANGS
#  * 'all': All of the above
#  * <csv>: Languages tags list to test
# Unset TEST_LANGS to test everything ('all')
# See: langs_dict()


"""Construct a dict of suites of languages to test.
{ '<suite name>' : <list or dict of language tags> }

ex.: { 'fetch' : {'en': 'English', 'fr': 'French'},
       'extra' : {'en': 'English', 'fr': 'French'} }
ex.: { 'environ' : ['en', 'fr'] }
"""
env = os.environ.get('TEST_LANGS')
if not env or env == 'all':
    langs = _fetch_langs()
    langs.update(_extra_langs())
elif env == 'fetch':
    langs = _fetch_langs()
elif env == 'extra':
    langs = _extra_langs()
else:
    env_langs = {l: l for l in env.split(',') if l}
    langs = env_langs


@pytest.mark.parametrize('lang', langs.keys(), ids=list(langs.values()))
def test_TTS(tmp_path, lang):
    """Test all supported languages and file save"""

    text = "This is a test"
    """Create output .mp3 file successfully"""
    for slow in (False, True):
        filename = tmp_path / 'test_{}_.mp3'.format(lang)
        # Create gTTS and save
        tts = gTTS(text=text, lang=lang, slow=slow)
        tts.save(filename)

        # Check if files created is > 2k
        assert filename.stat().st_size > 2000


def test_unsupported_language_check():
    """Raise ValueError on unsupported language (with language check)"""
    lang = 'xx'
    text = "Lorem ipsum"
    check = True
    with pytest.raises(ValueError):
        gTTS(text=text, lang=lang, lang_check=check)


def test_empty_string():
    """Raise AssertionError on empty string"""
    text = ""
    with pytest.raises(AssertionError):
        gTTS(text=text)


def test_no_text_parts(tmp_path):
    """Raises AssertionError on no content to send to API (no text_parts)"""
    text = "                                                                                                          ..,\n"
    with pytest.raises(AssertionError):
        filename = tmp_path / 'no_content.txt'
        tts = gTTS(text=text)
        tts.save(filename)


# Test write_to_fp()/save() cases not covered elsewhere in this file

def test_bad_fp_type():
    """Raise TypeError if fp is not a file-like object (no .write())"""
    # Create gTTS and save
    tts = gTTS(text='test')
    with pytest.raises(TypeError):
        tts.write_to_fp(5)


def test_save(tmp_path):
    """Save .mp3 file successfully"""
    filename = tmp_path / 'save.mp3'
    # Create gTTS and save
    tts = gTTS(text='test')
    tts.save(filename)

    # Check if file created is > 2k
    assert filename.stat().st_size > 2000


def test_get_urls():
    """Get request URLs list"""

    tts = gTTS(text='test', tld='com', lang='en-uk')
    urls = tts.get_urls()

    # Check the url
    r = urllib.parse.urlparse(urls[0])
    assert r.scheme == 'https'
    assert r.netloc == 'translate.google.com'
    assert r.path == '/translate_tts'
    assert 'test' in urllib.parse.parse_qs(r.query)['q']
    assert 'en-uk' in urllib.parse.parse_qs(r.query)['tl']


def test_msg():
    """Test gTTsError internal exception handling
    Set exception message successfully"""
    error1 = gTTSError('test')
    assert 'test' == error1.msg

    error2 = gTTSError()
    assert error2.msg is None


def test_infer_msg():
    """Infer message sucessfully based on context"""

    # Without response:

    # Bad TLD
    ttsTLD = Mock(tld='invalid')
    errorTLD = gTTSError(tts=ttsTLD)
    assert errorTLD.msg == "Failed to connect. Probable cause: Host 'https://translate.google.invalid/' is not reachable"

    # With response:

    # 403
    tts403 = Mock()
    response403 = Mock(status_code=403, reason='aaa')
    error403 = gTTSError(tts=tts403, response=response403)
    assert error403.msg == "403 (aaa) from TTS API. Probable cause: Bad token or upstream API changes"

    # 404 (and not lang_check)
    tts404 = Mock(lang='xx', lang_check=False)
    response404 = Mock(status_code=404, reason='bbb')
    error404 = gTTSError(tts=tts404, response=response404)
    assert error404.msg == "404 (bbb) from TTS API. Probable cause: Unsupported language 'xx'"

    # >= 500
    tts500 = Mock()
    response500 = Mock(status_code=500, reason='ccc')
    error500 = gTTSError(tts=tts500, response=response500)
    assert error500.msg == "500 (ccc) from TTS API. Probable cause: Uptream API error. Try again later."

    # Unknown (ex. 100)
    tts100 = Mock()
    response100 = Mock(status_code=100, reason='ddd')
    error100 = gTTSError(tts=tts100, response=response100)
    assert error100.msg == "100 (ddd) from TTS API. Probable cause: Unknown"


def test_WebRequest(tmp_path):
    """Test Web Requests"""

    text = "Lorem ipsum"

    """Raise gTTSError on unsupported language (without language check)"""
    lang = 'xx'
    check = False

    with pytest.raises(gTTSError):
        filename = tmp_path / 'xx.txt'
        # Create gTTS
        tts = gTTS(text=text, lang=lang, lang_check=check)
        tts.save(filename)


if __name__ == '__main__':
    pytest.main(['-x', __file__])
